/*
AUTHOR: ADRIAN DOBRA (ADOBRA@STAT.WASHINGTON.EDU)
*/
// gibbssampler.cpp: implementation of the CGibbsSampler class.
//
//////////////////////////////////////////////////////////////////////
#include "gibbssampler.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CGibbsSampler::CGibbsSampler()
{
	vslNewStream(&stream,BRNG,SEED);
	tauPrior = 1;
	deltaPrior = 2;
	myPI = 3.1415926535897931159979635;
	nMaxRegressors = 3; //this gives the maximum number of accepted regressors
	return;
}

CGibbsSampler::~CGibbsSampler()
{
	Cleanup();
	return;
}

void CGibbsSampler::Cleanup()
{
	vslDeleteStream(&stream);
	return;
}

int CGibbsSampler::MakeInteractionIndex(int p,int i,int j)
{
	int k;
	
	//make sure i < j
	if(i>j)
	{
		k = i;
		i = j;
		j = k;
	}
	//the index of each variable starts at zero
	i = i+1;
	j = j+1;
	return((int)(p+(i-1)*(p-(((double)i)/2.0))+j));
}

void CGibbsSampler::GetInteractionIndex(int ind,int p,int& vi,int& vj)
{
	int i,j;
    if(ind<=p)
	{
		vi = -1;
		vj = ind;
		return;
	}
	ind = ind-p;
	for(i=1;i<=p;i++)
	{
		j = ind-(i-1)*(p-(((double)i)/2.0));
		if((i<=j)&&(j<=p))
		{
			vi = i-1;
			vj = j-1;
			return;
		}
	}
	fprintf(stderr,"Fatal error :: GetInteractionIndex\n");
	exit(10);
	return;
} 

double* CGibbsSampler::NormalizeWeights(double* w,int nmax)
{
	int i;
	double* cumw = new double[nmax+1];

	//the weights are in increasing order
	//divide by the largest one, i.e. substract its log
	for(i=0;i<nmax;i++)
	{
		w[i] -= w[nmax-1];
	}
 
	for(i=0;i<nmax;i++)
	{
		w[i] = AdjustDouble(exp(w[i]));
	}

    cumw[0] = 0;
	for(i=1;i<=nmax;i++)
	{
		cumw[i] = cumw[i-1]+w[i-1];
    }
    for(i=1;i<=nmax;i++)
	{
		cumw[i] /= cumw[nmax];
    }
    return cumw;
}

int CGibbsSampler::WeightedSampling(int maxk,double* weights)
{
	double s;
	int i;
	int q1 = 0;
	int q2 = maxk;

	vdRngUniform(METHOD,stream,1,&s,0,1);
	while(q1+1!=q2)
	{
		int q = (q1+q2)/2;
		if(weights[q]<s)
		{
			q1 = q;
		}
		else
		{
			q2 = q;
		}
	}
	return(q1);  
}

int CGibbsSampler::SelectEquation(int NumberOfGenes,CEquations* eq)
{
	int i,j;
	int equation;
	int chosenReg = -1;
	double delta;

	double* weights = new double[NumberOfGenes];
	int* ids = new int[NumberOfGenes];

	int nmax = eq->harvest(ids,weights);
	if(0==nmax)
	{
		return(-1);
	}
	
	//normalize the weights
	int chosenNeighbor = 0;
	double* cumw = NormalizeWeights(weights,nmax);
	chosenNeighbor = WeightedSampling(nmax,cumw);
	delete[] cumw; cumw = NULL;

	chosenReg = ids[chosenNeighbor];

	//clean memory
	delete[] weights; weights = NULL;
	delete[] ids; ids = NULL;
	return(chosenReg);
}

double CGibbsSampler::wishartPosterior(int npred, CData& Data)
{
	int i,j,k;
	static int n = Data.SampleSize;
	double nu = tauPrior;
	double s;
	double logPost = 0.0;

	for(i=0;i<n;i++) {
	  nu += Data.Y[i] * Data.Y[i];
	}

	double* MX = new double[npred * npred];
	memset(MX,0,npred * npred *sizeof(double));
	for(i=0;i<npred;i++) {
		for(j=i;j<npred;j++) {
			s = 0.0;
			for(k=0;k<n;k++) {
				s += Data.X[k][i]*Data.X[k][j];
			}
         
			if(i==j) {
				MX[i*npred + i] = s+tauPrior;
			} else {
				MX[i * npred + j] = MX[j * npred +i] = s;
			}
		}
	}
	
	//calculate the determinant of M
	double logDetM = getlogdet(npred, MX);
	
	//calculate the inverse of M
	//this inverse will be stored in M
	if(!Solve(npred,MX)) {
		return(DBL_MAX);
	} 

	double* Zy = new double[npred];
	memset(Zy,0,npred*sizeof(double));
	for(i=0;i<npred;i++) {
		s = 0.0;
		for(k=0;k<n;k++) {
			s += Data.X[k][i]*Data.Y[k] ;
		}
		Zy[i] = s;
	}

	s = 0.0;
	for(i=0;i<npred;i++) {
		for(j=0;j<npred;j++) {
			s += Zy[i]*MX[i * npred + j]*Zy[j];
		}
	}
	//finished calculating nu
	nu -= s;

	logPost -= n*log(myPI)/2.0;
	logPost += (deltaPrior+2*npred)*log(tauPrior)/2.0;
	logPost -= (logDetM/2.0);
	logPost += lgamma((n+deltaPrior+npred)/2.0);
	logPost -= lgamma((deltaPrior+npred)/2.0);
	logPost -= (n+deltaPrior+npred)*log(nu)/2.0; 

	//clean memory
	delete[] MX; MX = NULL;
	delete[] Zy; Zy = NULL;
	return(logPost);
}

double CGibbsSampler::ymarginalZeroPredictors(int TargetGene, CData& Data)
{
	int i,l;
	double s = 0.0;
	for(l=0;l<Data.SampleSize;l++) {
		s += (Data.data[l][TargetGene]*Data.data[l][TargetGene]);
	}
	
	double logPost = 0.0;
	logPost -= Data.SampleSize*log(myPI)/2.0;
	logPost += deltaPrior*log(tauPrior)/2.0;
	logPost += lgamma((Data.SampleSize+deltaPrior)/2.0);
	logPost -= lgamma(deltaPrior/2.0);
	logPost -= (Data.SampleSize+deltaPrior)*log(tauPrior+s)/2.0;
    
	return(logPost);
}

double CGibbsSampler::calculateLogPost(int TargetGene, set<int>& ExplanatoryGenes, CData& Data)
{
	int i,j;
	double logPost = 0.0;
	//////////////////////////////////
	//initialize the Y and X vectors//
	//////////////////////////////////
	for(j=0;j<Data.SampleSize;j++)
	{
		Data.Y[j] = Data.data[j][TargetGene];
	}
	
	set<int>::iterator it;
	for (i = 0, it = ExplanatoryGenes.begin();it!=ExplanatoryGenes.end();(it++,i++))
	{
		int gene = *it;
		if(i>=Data.ncolsX)
		{
			fprintf(stderr,"Data X matrix has only %d columns. Attempted to use more.\n",Data.ncolsX);
			exit(1);
		}
		if(gene<Data.NumberOfGenes)
		{
			for(j=0;j<Data.SampleSize;j++)
			{ 
				Data.X[j][i] = Data.data[j][*it];
			}
		}
		else //this is an interaction term
		{
			int vi, vj;
			GetInteractionIndex(gene,Data.NumberOfGenes-1,vi,vj);
			//create the interaction term
			double sM = 0.0;
			double sD = 0.0;
			for(j=0;j<Data.SampleSize;j++)
			{
				Data.X[j][i] = Data.data[j][vi]*Data.data[j][vj];
				sM += Data.X[j][i];
				sD += pow(Data.X[j][i],2);
			}
			sM = sM/((double)Data.SampleSize);
			sD = sD/((double)Data.SampleSize);
			sD = sqrt(sD-pow(sM,2));
			for(j=0;j<Data.SampleSize;j++)
			{
				Data.X[j][i] = (Data.X[j][i]-sM)/sD;
			}
		}
	}
	logPost = wishartPosterior(ExplanatoryGenes.size(), Data);
	return(logPost);
}

int CGibbsSampler::ChooseNextGene(vector<int>& PossiblePredictors)
{
	int index;
	if(0==PossiblePredictors.size())
	{
		return(-1);
	}
	viRngUniform(METHOD,stream,1,&index,0,PossiblePredictors.size());
	int selectedGene = PossiblePredictors[index];
	PossiblePredictors.erase(PossiblePredictors.begin() + index);
	return selectedGene;
}

int CGibbsSampler::ChooseNextGeneSet(set<int>& PossiblePredictors)
{
	int i;
	int index;
	if(0==PossiblePredictors.size())
	{
		return(-1);
	}
	viRngUniform(METHOD,stream,1,&index,0,PossiblePredictors.size());
	set<int>::iterator it;
	i = 0;
	for(it=PossiblePredictors.begin();it!=PossiblePredictors.end();it++)
	{
		if(i==index)
		{
			return(*it);
		}
		i++;
	}
	return(-1);
}

int CGibbsSampler::convergenceReg(int lenv,double* v1,double* v2)
{
	int i;
	double s = 0.0;
	
	for(i=0;i<lenv;i++)
	{
		s+=fabs(v1[i]-v2[i]);
		//printf("%.3lf\t%.3lf\n",v1[i],v2[i]);
	}
	printf("diff = %.5lf\n",s);
	if(s<0.01) return(1);
	return(0);
}

double CGibbsSampler::hybridReg(int targetGene,CData& Data,double* probs,FILE* modelsfile)
{
	int i;
	int gene;
	double currentMarginalLik;
	double delta, ratio;
	double* oldprobs = new double[Data.NumberOfGenes];
	double* newprobs = new double[Data.NumberOfGenes];
	double u;
	
	int nFirstOrderEffects[Data.NumberOfGenes];
	for(i=0;i<Data.NumberOfGenes;i++)
	{
		nFirstOrderEffects[i] = 0;
	}
	
	set<int> VarInModel;
	set<int> emptyVarInModel;
	double emptyregpost = ymarginalZeroPredictors(targetGene, Data);
	currentMarginalLik = emptyregpost;
	
	//this stores a sorted list of all equations
	CEquations* eq = new CEquations;
	CRegression* reglist = new CRegression;
	
	for (i=0;i<Data.NumberOfGenes;i++)
	{
		if(i!=targetGene)
		{
			newprobs[i] = 1.0;
		}
		else
		{
			newprobs[i] = 0.0;
		}
	}
	
	set<int> allPredictors;
    for(i=0;i<Data.NumberOfGenes;i++)
	{
		if(i!=targetGene)
		{
			allPredictors.insert(i);
		}
	}
	
	mGeneList.clear();
	for(i=0;i<Data.NumberOfGenes;i++)
	{
		mGeneList.push_back(i);
	}
	mGeneList.erase(mGeneList.begin()+targetGene);

	int nPredictors = Data.NumberOfGenes-1;
	//storage for the candidates
	int nMaxCandidates = (int)(10*(nMaxRegressors/2.0)*((double)(nMaxRegressors+3)))+1;
	int* Candidates = new int[nMaxCandidates];
	double* Score = new double[nMaxCandidates];
	int nCandidates = 0;

	reglist->AddReg(currentMarginalLik,VarInModel);
	int counter = 0;
	do
	{
		printf("Step [%d]\n",++counter);
		for(i=0;i<Data.NumberOfGenes;i++)
		{
			oldprobs[i] = newprobs[i];
		}

		for(int myreplicate=0;myreplicate<nChainReplicates;myreplicate++)
		{
			for(i=0;i<Data.NumberOfGenes;i++)
			{
				nFirstOrderEffects[i] = 0;
			}
			
			VarInModel = reglist->Next->Vars;
			currentMarginalLik = reglist->Next->Weight;
			set<int> Predictors = allPredictors;
			set<int>::iterator it;
			for(it=VarInModel.begin();it!=VarInModel.end();it++)
			{
				gene = *it;
				if(gene<Data.NumberOfGenes)
				{
					Predictors.erase(*it);
				}
				else
				{
					int vi, vj;
					GetInteractionIndex(gene,Data.NumberOfGenes-1,vi,vj);
					if(vi>=0)
					{
						nFirstOrderEffects[vi] += 1;
					}
					else
					{
						fprintf(stderr,"vi = %d PROGRAM HAS TO EXIT\n",vi);
						exit(10);
					}
					if(vj>=0)
					{
						nFirstOrderEffects[vj] += 1;
					}
					else
					{
						fprintf(stderr,"vj = %d PROGRAM HAS TO EXIT\n",vj);
						exit(10);
					}
				}
			}

			for(int myiter=0;myiter<nChainIterations;myiter++)
			{
				//GLOBAL SHOTGUN PART
				nCandidates = 0;
				//record the current model as a candidate
				eq->Add(nCandidates,0);
				Candidates[nCandidates] = -1;
				Score[nCandidates] = 0;
				nCandidates++; 

				//main effects
				set<int> shotgunPredictors = allPredictors;
				while(!shotgunPredictors.empty())
				{
					gene = ChooseNextGeneSet(shotgunPredictors);
					if(gene<0) break; //just in case
					shotgunPredictors.erase(gene);
					
					int overflowFlag = -1;
					//test if the variable is currently in the model
					if(VarInModel.find(gene) == VarInModel.end())
					{
						overflowFlag = 0;
						//add the variable to the model
						VarInModel.insert(gene);
						if(VarInModel.size() > nMaxRegressors)
						{
							overflowFlag = 1;
						}
						else
						{
							delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
						}
						//delete the variable from the model
						VarInModel.erase(gene);
					}  
					else if(VarInModel.find(gene) != VarInModel.end())
					{
						//we can delete a variable only if it does
						//not have any dependencies
						if(0==nFirstOrderEffects[gene])
						{
							overflowFlag = 0;

							//take the variable out of the model
							VarInModel.erase(gene);
							if(VarInModel.empty())
							{
								delta = ymarginalZeroPredictors(targetGene, Data)-currentMarginalLik;
							}
							else
							{
								delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
							}

							//put the variable back in the model
							VarInModel.insert(gene);
						}
					}

					if(0==overflowFlag)
					{
						if(nCandidates<nMaxCandidates)
						{
							Candidates[nCandidates] = gene;
							eq->Add(nCandidates,delta);
							Score[nCandidates] = delta;
							nCandidates++;
						}
						else
						{
							fprintf(stderr,"nCandidates = %d and maximum number is %d PROGRAM EXITS\n",nCandidates,nMaxCandidates);
							exit(1);
						}
					}
				}
				
				//printf("got here 1\n");
				//interaction terms
				set<int>::iterator itI;
				set<int>::iterator itJ;
				for(itI=VarInModel.begin();itI!=VarInModel.end();itI++)
				{
					int geneI = *itI;
					if(geneI<Data.NumberOfGenes)
					{
						for(itJ=itI;itJ!=VarInModel.end();itJ++)
						{
							int geneJ = *itJ;
							if(geneJ<Data.NumberOfGenes)
							{
								gene = MakeInteractionIndex(Data.NumberOfGenes-1,geneI,geneJ);
								int overflowFlag = -1;
								if(VarInModel.find(gene) == VarInModel.end())
								{
									overflowFlag = 0;
									//add the variable to the model
									VarInModel.insert(gene);
									if(VarInModel.size() > nMaxRegressors)
									{
										overflowFlag = 1;
									}
									else
									{
										delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
									}
									//delete the variable from the model
									VarInModel.erase(gene);
								}  
								else if(VarInModel.find(gene) != VarInModel.end())
								{
									overflowFlag = 0;

									//take the variable out of the model
									VarInModel.erase(gene);
									if(VarInModel.empty())
									{
										delta = ymarginalZeroPredictors(targetGene, Data)-currentMarginalLik;
									}
									else
									{
										delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
									}

									//put the variable back in the model
									VarInModel.insert(gene);
								}
								if(0==overflowFlag)
								{	
									if(nCandidates<nMaxCandidates)
									{
										Candidates[nCandidates] = gene;
										eq->Add(nCandidates,delta);
										Score[nCandidates] = delta;
										nCandidates++;
									}
									else
									{
										fprintf(stderr,"nCandidates = %d and maximum number is %d PROGRAM EXITS\n",nCandidates,nMaxCandidates);
										exit(1);
									}        
								}
							}
						}
					}
				}
				//printf("got here 2\n");
				
				if(1==nCandidates)
				{
					eq->DeleteAll();
					continue;
				}
				int chosenCandidate = SelectEquation(nCandidates,eq);
				//printf("chosenCandidate = %d :: nCandidates = %d\n",chosenCandidate,nCandidates);
				gene = Candidates[chosenCandidate];
				eq->DeleteAll();

				//printf("gene = %d\n",gene);
				if(gene>=0)
				{
					if(VarInModel.find(gene) == VarInModel.end())
					{
						VarInModel.insert(gene);
						if(gene<Data.NumberOfGenes)
						{
							Predictors.erase(gene);
						}
						else
						{
							int vi, vj;
							GetInteractionIndex(gene,Data.NumberOfGenes-1,vi,vj);
							//fprintf(stderr,"gene = %d :: vi = %d :: vj = %d\n",gene,vi,vj);
							nFirstOrderEffects[vi] += 1;
							nFirstOrderEffects[vj] += 1;
						}
					}
					else
					{
						VarInModel.erase(gene);
						if(gene<Data.NumberOfGenes)
						{
							Predictors.insert(gene);
						}
						else
						{
							int vi, vj;
							GetInteractionIndex(gene,Data.NumberOfGenes-1,vi,vj);
							//fprintf(stderr,"gene = %d :: vi = %d :: vj = %d\n",gene,vi,vj);
							nFirstOrderEffects[vi] -= 1;
							nFirstOrderEffects[vj] -= 1;
							if(nFirstOrderEffects[vi]<0)
							{
								fprintf(stderr,"Fatal error :: nFirstOrderEffects[%d] = %d\n",
									    vi,nFirstOrderEffects[vi]);
								exit(1);
							}
							if(nFirstOrderEffects[vj]<0)
							{
								fprintf(stderr,"Fatal error :: nFirstOrderEffects[%d] = %d\n",
									    vj,nFirstOrderEffects[vj]);
								exit(1);
							}
						}
					}
					//printf("got here 2.5\n");
					if(VarInModel.empty())
					{
						currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
					}
					else
					{
						currentMarginalLik = calculateLogPost(targetGene, VarInModel, Data);
					}
					reglist->AddReg(currentMarginalLik,VarInModel);
				}
				//printf("got here 3\n");
			}
			reglist->Delete();
		}
		
		reglist->NormalizeWeights();
		reglist->MakeCovProbs(Data.NumberOfGenes,newprobs);
	}while(!convergenceReg(Data.NumberOfGenes,newprobs,oldprobs));

	for(i=0;i<Data.NumberOfGenes;i++)
	{
		probs[i] = newprobs[i];
	}
	reglist->SaveAll(nMaxRegressors,modelsfile);
	double bestPost = reglist->Next->Weight;
	reglist->DeleteAll();

	//clean memory
	delete[] oldprobs;
	delete[] newprobs;
	delete eq;
	delete reglist;
	delete[] Score; Score = NULL;
	delete[] Candidates; Candidates = NULL;
	return(bestPost);
}