/*
AUTHOR: ADRIAN DOBRA (ADOBRA@STAT.WASHINGTON.EDU)
*/
// gibbssampler.cpp: implementation of the CGibbsSampler class.
//
//////////////////////////////////////////////////////////////////////
#include "gibbssampler.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CGibbsSampler::CGibbsSampler()
{
	vslNewStream(&stream,BRNG,SEED);
	tauPrior = 1;
	deltaPrior = 2;
	myPI = 3.1415926535897931159979635;
	nMaxRegressors = 5; //this gives the maximum number of accepted regressors
	return;
}

CGibbsSampler::~CGibbsSampler()
{
	Cleanup();
	return;
}

void CGibbsSampler::Cleanup()
{
	vslDeleteStream(&stream);
	return;
}

double* CGibbsSampler::NormalizeWeights(double* w,int nmax)
{
	int i;
	double* cumw = new double[nmax+1];

	//the weights are in increasing order
	//divide by the largest one, i.e. substract its log
	for(i=0;i<nmax;i++)
	{
		w[i] -= w[nmax-1];
	}
 
	for(i=0;i<nmax;i++)
	{
		w[i] = AdjustDouble(exp(w[i]));
	}

    cumw[0] = 0;
	for(i=1;i<=nmax;i++)
	{
		cumw[i] = cumw[i-1]+w[i-1];
    }
    for(i=1;i<=nmax;i++)
	{
		cumw[i] /= cumw[nmax];
    }
    return cumw;
}

int CGibbsSampler::WeightedSampling(int maxk,double* weights)
{
	double s;
	int i;
	int q1 = 0;
	int q2 = maxk;

	vdRngUniform(METHOD,stream,1,&s,0,1);
	while(q1+1!=q2)
	{
		int q = (q1+q2)/2;
		if(weights[q]<s)
		{
			q1 = q;
		}
		else
		{
			q2 = q;
		}
	}
	return(q1);  
}

int CGibbsSampler::SelectEquation(int NumberOfGenes,CEquations* eq)
{
	int i,j;
	int equation;
	int chosenReg = -1;
	double delta;

	double* weights = new double[NumberOfGenes];
	int* ids = new int[NumberOfGenes];

	int nmax = eq->harvest(ids,weights);
	if(0==nmax)
	{
		return(-1);
	}
	
	//normalize the weights
	int chosenNeighbor = 0;
	double* cumw = NormalizeWeights(weights,nmax);
	chosenNeighbor = WeightedSampling(nmax,cumw);
	delete[] cumw; cumw = NULL;

	chosenReg = ids[chosenNeighbor];

	//clean memory
	delete[] weights; weights = NULL;
	delete[] ids; ids = NULL;
	return(chosenReg);
}

double CGibbsSampler::wishartPosterior(int npred, CData& Data)
{
	int i,j,k;
	static int n = Data.SampleSize;
	double nu = tauPrior;
	double s;
	double logPost = 0.0;

	for(i=0;i<n;i++) {
	  nu += Data.Y[i] * Data.Y[i];
	}

	double* MX = new double[npred * npred];
	memset(MX,0,npred * npred *sizeof(double));
	for(i=0;i<npred;i++) {
		for(j=i;j<npred;j++) {
			s = 0.0;
			for(k=0;k<n;k++) {
				s += Data.X[k][i]*Data.X[k][j];
			}
         
			if(i==j) {
				MX[i*npred + i] = s+tauPrior;
			} else {
				MX[i * npred + j] = MX[j * npred +i] = s;
			}
		}
	}
	
	//calculate the determinant of M
	double logDetM = getlogdet(npred, MX);
	
	//calculate the inverse of M
	//this inverse will be stored in M
	if(!Solve(npred,MX)) {
		return(DBL_MAX);
	} 

	double* Zy = new double[npred];
	memset(Zy,0,npred*sizeof(double));
	for(i=0;i<npred;i++) {
		s = 0.0;
		for(k=0;k<n;k++) {
			s += Data.X[k][i]*Data.Y[k] ;
		}
		Zy[i] = s;
	}

	s = 0.0;
	for(i=0;i<npred;i++) {
		for(j=0;j<npred;j++) {
			s += Zy[i]*MX[i * npred + j]*Zy[j];
		}
	}
	//finished calculating nu
	nu -= s;

	logPost -= n*log(myPI)/2.0;
	logPost += (deltaPrior+2*npred)*log(tauPrior)/2.0;
	logPost -= (logDetM/2.0);
	logPost += lgamma((n+deltaPrior+npred)/2.0);
	logPost -= lgamma((deltaPrior+npred)/2.0);
	logPost -= (n+deltaPrior+npred)*log(nu)/2.0; 

	//clean memory
	delete[] MX; MX = NULL;
	delete[] Zy; Zy = NULL;
	return(logPost);
}

double CGibbsSampler::ymarginalZeroPredictors(int TargetGene, CData& Data)
{
	int i,l;
	double s = 0.0;
	for(l=0;l<Data.SampleSize;l++) {
		s += (Data.data[l][TargetGene]*Data.data[l][TargetGene]);
	}
	
	double logPost = 0.0;
	logPost -= Data.SampleSize*log(myPI)/2.0;
	logPost += deltaPrior*log(tauPrior)/2.0;
	logPost += lgamma((Data.SampleSize+deltaPrior)/2.0);
	logPost -= lgamma(deltaPrior/2.0);
	logPost -= (Data.SampleSize+deltaPrior)*log(tauPrior+s)/2.0;
    
	return(logPost);
}

double CGibbsSampler::calculateLogPost(int TargetGene, set<int>& ExplanatoryGenes, CData& Data) {
	int i,j;
	double logPost = 0.0;
	//////////////////////////////////
	//initialize the Y and X vectors//
	//////////////////////////////////
	for(j=0;j<Data.SampleSize;j++) {
		Data.Y[j] = Data.data[j][TargetGene];
	}
	
	set<int>::iterator it;
	for (i = 0, it = ExplanatoryGenes.begin();it!=ExplanatoryGenes.end();(it++,i++)) {
		for(j=0;j<Data.SampleSize;j++) { 
			Data.X[j][i] = Data.data[j][*it];
		}
	}

	logPost = wishartPosterior(ExplanatoryGenes.size(), Data);
	return(logPost);
}

int CGibbsSampler::ChooseNextGene(vector<int>& PossiblePredictors)
{
	int index;
	viRngUniform(METHOD,stream,1,&index,0,PossiblePredictors.size());
	int selectedGene = PossiblePredictors[index];
	PossiblePredictors.erase(PossiblePredictors.begin() + index);
	return selectedGene;
}

double CGibbsSampler::metropolisReg(int targetGene,CData& Data)
{
	int i,j,k;
	int gene;
	double currentMarginalLik;
	double ratio;
	double pData;
	
	set<int> VarInModel;
	
	mGeneList.clear();
	for (i = 0; i < Data.NumberOfGenes; i++)
	{
		mGeneList.push_back(i);
	}
	mGeneList.erase(mGeneList.begin() + targetGene);
         
	//generate a starting model
	int nPredictors = Data.NumberOfGenes-1;
	vector<int> Predictors(mGeneList);
        
	//calculate initial model probabilities
	if(VarInModel.empty())
	{
		currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
	}
	else
	{
		currentMarginalLik= calculateLogPost(targetGene, VarInModel, Data);
	}

	double bestPost = -9999999999999999.0;
	int    myiter = 0;
	while(myiter<nChainIterations*nPredictors)
	{
	   //fprintf(stderr,"Iteration = %d\n",myiter+1);
	   Predictors = mGeneList;
	   gene = ChooseNextGene(Predictors);
	      
		int overflowFlag = -1;
	   //test if the variable is currently in the model
	   if (VarInModel.find(gene) == VarInModel.end())
	   { 
		  //the variable is not in the model
	      overflowFlag = 0;
	      //add the variable to the model
	      VarInModel.insert(gene);
	      if(VarInModel.size() > nMaxRegressors)
		  {
	         overflowFlag = 1;
	      }
		  else
		  {
			 pData = calculateLogPost(targetGene, VarInModel, Data);
             ratio = pData-currentMarginalLik;
	      }
	      //delete the variable from the model
	      VarInModel.erase(gene);
	   }  
	   else if(VarInModel.find(gene) != VarInModel.end())
	   {
	      overflowFlag = 0;

	      //take the variable out of the model
	      VarInModel.erase(gene);
	      if(VarInModel.empty())
		  {
	         pData = ymarginalZeroPredictors(targetGene, Data);
	      }
		  else
		  {
		     pData = calculateLogPost(targetGene, VarInModel, Data);
	      }

	      //put the variable back in the model
	      VarInModel.insert(gene);
		  ratio =  pData-currentMarginalLik;
	   }

	   if(0==overflowFlag)
	   {
	      ratio = exp(ratio);
		  if(ratio>1)
	      {
	         ratio = 1;
	      }

	      double u;
		  vdRngUniform(METHOD,stream,1,&u,0,1);
		  
	      if(u<ratio)
	      {  //move is accepted
		     currentMarginalLik = pData;
		     if(VarInModel.find(gene) == VarInModel.end())
			 {
				VarInModel.insert(gene);
			 }
			 else  
		     {
				VarInModel.erase(gene);
			 }
		  }

		  double s = currentMarginalLik;        
		  if(s>bestPost)
		  {
		     bestPost = s;
		  }
	   }
	   myiter++;
	}
    return(bestPost);
}

double CGibbsSampler::gibbsSamplingReg(int targetGene,CData& Data)
{
	int i,j,k;
	int gene;
	double currentMarginalLik;
	double ratio;
	double pDataZero, pDataOne;
	
	set<int> VarInModel;
	
	mGeneList.clear();
	for(i = 0; i < Data.NumberOfGenes; i++)
	{
		mGeneList.push_back(i);
	}
	mGeneList.erase(mGeneList.begin() + targetGene);
         
	//generate a starting model
	int nPredictors = Data.NumberOfGenes-1;
	vector<int> Predictors(mGeneList);
        
	//calculate initial model probabilities
	if(VarInModel.empty())
	{
		currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
	}
	else
	{
		currentMarginalLik= calculateLogPost(targetGene, VarInModel, Data);
	}

	double bestPost = -9999999999999999.0;
	int    myiter = 0;
	while(myiter<nChainIterations)
	{
	   //fprintf(stderr,"Iteration = %d\n",myiter+1);
	   Predictors = mGeneList;
	   for(int ii=0;ii<nPredictors;ii++)
	   {
	      gene = ChooseNextGene(Predictors);
	      int overflowFlag = -1;
	      //test if the variable is currently in the model
	      if (VarInModel.find(gene) == VarInModel.end())
              {
	         overflowFlag = 0;
		 pDataZero = currentMarginalLik;
		 //add the variable to the model
		 VarInModel.insert(gene);
		 if(VarInModel.size() > nMaxRegressors)
                 {
		    overflowFlag = 1;
		 }
                 else
                 {
		    pDataOne = calculateLogPost(targetGene, VarInModel, Data);
		 }
		 //delete the variable from the model
		 VarInModel.erase(gene);
	      }  
              else if(VarInModel.find(gene) != VarInModel.end())
              {
	         overflowFlag = 0;
		 pDataOne = currentMarginalLik;

		 //take the variable out of the model
		 VarInModel.erase(gene);
		 if(VarInModel.empty())
                 {
		    pDataZero = ymarginalZeroPredictors(targetGene, Data);
		 }
                 else
                 {
		    pDataZero = calculateLogPost(targetGene, VarInModel, Data);
		 }

		 //put the variable back in the model
		 VarInModel.insert(gene);
	      }

	      if(0==overflowFlag)
              {
	         ratio = pDataZero-pDataOne;
		 ratio = 1/(1+exp(ratio));

		 double u;
		 vdRngUniform(METHOD,stream,1,&u,0,1);
		 
		 if(u<ratio)
                 { //variable should be in the model
		    if(VarInModel.find(gene) == VarInModel.end())
                    {
		       VarInModel.insert(gene);
		       currentMarginalLik = pDataOne;
		    }
		 }
                 else
                 {//variable should not be in the model
		    if(VarInModel.find(gene) != VarInModel.end())
                    {
		       VarInModel.erase(gene);
		       currentMarginalLik = pDataZero;
		    }
		 }
	      }
              
              double s = currentMarginalLik;        
              if(s>bestPost)
	      {
	         bestPost = s;
	      }
	   }

	   myiter++;
	}
   
        return(bestPost);
}

double CGibbsSampler::shotgunReg(int targetGene,CData& Data)
{
	int i,j,k;
	int gene;
	double currentMarginalLik;
    double delta;
	
	set<int> VarInModel;

	//this stores a sorted list of all equations
	CEquations* eq = new CEquations();
	
	mGeneList.clear();
	for (i=0;i<Data.NumberOfGenes;i++)
	{
		mGeneList.push_back(i);
	}
	mGeneList.erase(mGeneList.begin() + targetGene);
         
	//generate a starting model
	int nPredictors = Data.NumberOfGenes-1;
	vector<int> Predictors(mGeneList);

        //storage for the candidates
        int nCandidates = 0;
        int* Candidates = new int[nPredictors+1];
        double* Score = new double[nPredictors+1];
        
	//calculate initial model probabilities
	if(VarInModel.empty()) {
		currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
	} else {
		currentMarginalLik= calculateLogPost(targetGene, VarInModel, Data);
	}

    double bestPost = -9999999999999999.0;
	int    myiter = 0;
	while(myiter<nChainIterations)
        {
           nCandidates = 0;
           //record the current model as a candidate
           eq->Add(nCandidates,0);
           Candidates[nCandidates] = -1;
           Score[nCandidates] = 0;
           nCandidates++; 

	   //fprintf(stderr,"Iteration = %d\n",myiter+1);
	   Predictors = mGeneList;
	   for(int ii=0;ii<nPredictors;ii++)
           {
	      gene = ChooseNextGene(Predictors);
	      int overflowFlag = -1;
	      //test if the variable is currently in the model
	      if (VarInModel.find(gene) == VarInModel.end())
              {
	         overflowFlag = 0;
		 //add the variable to the model
		 VarInModel.insert(gene);
		 if(VarInModel.size() > nMaxRegressors)
                 {
		    overflowFlag = 1;
		 }
                 else
                 {
		    delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
		 }
		 //delete the variable from the model
		 VarInModel.erase(gene);
	      }  
              else if(VarInModel.find(gene) != VarInModel.end())
              {
	         overflowFlag = 0;

		 //take the variable out of the model
		 VarInModel.erase(gene);
		 if(VarInModel.empty())
                 {
		    delta = ymarginalZeroPredictors(targetGene, Data)-currentMarginalLik;
		 }
                 else
                 {
		    delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
		 }

		 //put the variable back in the model
		 VarInModel.insert(gene);
	      }

	      if(0==overflowFlag)
              {
                 Candidates[nCandidates] = gene;
                 eq->Add(nCandidates,delta);
                 Score[nCandidates] = delta;
                 nCandidates++;           
              }
	   }

           int chosenCandidate = SelectEquation(nCandidates,eq);
           eq->DeleteAll();

           if(chosenCandidate>=1)
           {
              gene = Candidates[chosenCandidate];

              if(VarInModel.find(gene) == VarInModel.end())
              {
                 VarInModel.insert(gene);  
	      }
              else
	      {
                 VarInModel.erase(gene);
	      }
              
               if(VarInModel.empty())
               {
	          currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
	       }
               else
               {
	          currentMarginalLik = calculateLogPost(targetGene, VarInModel, Data);
	       }
	   }
                
           double s = currentMarginalLik;        
           if(s>bestPost)
	   {
	      bestPost = s;
	   }        
           
	   myiter++;
	}
        

	//clean memory
	delete eq;
	delete[] Score; Score = NULL;
	delete[] Candidates; Candidates = NULL;
	return(bestPost);
}

int CGibbsSampler::convergenceReg(int lenv,double* v1,double* v2)
{
	int i;
	double s = 0.0;
	
	for(i=0;i<lenv;i++)
	{
		s+=fabs(v1[i]-v2[i]);
	}
	if(s<0.1) return(1);
	return(0);
}

int CGibbsSampler::randomWeightedSelect(double* probs,set<int>& Predictors)
{
	double w = 0.0;
	int gene = -1;
	double u;
	set<int>::iterator it;
	
	double sumprobs = 0.0;
	for(it=Predictors.begin();it!=Predictors.end();it++)
	{
		gene = *it;
		sumprobs += probs[gene];
	}
	
	vdRngUniform(METHOD,stream,1,&u,0,1);	
	for(it=Predictors.begin();it!=Predictors.end();it++)
	{
		gene = *it;
		w += (probs[gene]/sumprobs);
		if(w>u)
		{
			break;
		}
	}
	
	return(gene);
}

void CGibbsSampler::hybridReg(int targetGene,CData& Data,double* probs,FILE* modelsfile)
{
	int i;
	int gene;
	double currentMarginalLik;
	double delta;
	double newprobs[Data.NumberOfGenes];
	int nPredictors = Data.NumberOfGenes-1;
	set<int> Predictors;
	
	set<int> VarInModel;
	//this stores a sorted list of all equations
	CEquations* eq = new CEquations;
	CRegression* reglist = new CRegression;
	
	for (i=0;i<Data.NumberOfGenes;i++)
	{
		if(i!=targetGene)
		{
			Predictors.insert(i);
			newprobs[i] = 1.0/((double)(Data.NumberOfGenes-1));
		}
		else
		{
			newprobs[i] = 0.0;
		}
	}

	//storage for the candidates
	int nCandidates = 0;
	int* Candidates = new int[nMaxRegressors+2];
	double* Score = new double[nMaxRegressors+2];

	//always start from the model with no predictors
	//calculate initial model probabilities
	if(VarInModel.empty())
	{
		currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
	}
	else 
	{
		currentMarginalLik= calculateLogPost(targetGene, VarInModel, Data);
	}
	reglist->AddReg(currentMarginalLik,VarInModel);

	int counter = 0;
    do
	{
		printf("Step [%d]\n",++counter);
		for(i=0;i<Data.NumberOfGenes;i++)
		{
			probs[i] = newprobs[i];
		}
	
		for(int myiter=0;myiter<nChainIterations*nPredictors;myiter++)
		{
			nCandidates = 0;
			if(VarInModel.size()<nMaxRegressors)//add the current model
			{
				eq->Add(nCandidates,0);
				Candidates[nCandidates] = -1;
				Score[nCandidates] = 0;
				nCandidates++;
				
				//select one variable to add
				gene = randomWeightedSelect(probs,Predictors);
				if(gene>=0)
				{
					VarInModel.insert(gene);
					delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
					VarInModel.erase(gene);
					Candidates[nCandidates] = gene;
					eq->Add(nCandidates,delta);
					Score[nCandidates] = delta;
					nCandidates++;
				}
			}
			//take one variable out of the current regression
			//and record the resulting models
			set<int>::iterator it;
			set<int> copyVarInModel = VarInModel; 
			for(it=copyVarInModel.begin();it!=copyVarInModel.end();it++)
			{
				gene = *it;
				VarInModel.erase(gene);
				if(VarInModel.empty())
				{
					delta = ymarginalZeroPredictors(targetGene, Data)-currentMarginalLik;
				}
				else 
				{
					delta = calculateLogPost(targetGene, VarInModel, Data)-currentMarginalLik;
				}
				Candidates[nCandidates] = gene;
				eq->Add(nCandidates,delta);
				Score[nCandidates] = delta;
				nCandidates++;    
				VarInModel.insert(gene);
			}
			
			//select a variable to add or delete
			int chosenCandidate = SelectEquation(nCandidates,eq);
			gene = Candidates[chosenCandidate];
			eq->DeleteAll();
			if(gene>=0)
			{
				if(VarInModel.find(gene) == VarInModel.end())
				{
					VarInModel.insert(gene);
					Predictors.erase(gene);
				}
				else
				{
					VarInModel.erase(gene);
					Predictors.insert(gene);
				}
				if(VarInModel.empty())
				{
					currentMarginalLik = ymarginalZeroPredictors(targetGene, Data);
				}
				else
				{
					currentMarginalLik = calculateLogPost(targetGene, VarInModel, Data);
				}
				reglist->AddReg(currentMarginalLik,VarInModel);
			}
		}
		reglist->NormalizeWeights();
		reglist->MakeCovProbs(Data.NumberOfGenes,newprobs);
	}while(!convergenceReg(Data.NumberOfGenes,newprobs,probs));

	for(i=0;i<Data.NumberOfGenes;i++)
	{
		probs[i] = newprobs[i];
	}
	reglist->SaveAll(nMaxRegressors,modelsfile);
	reglist->DeleteAll();

	//clean memory
	delete eq;
	delete reglist;
	delete[] Score; Score = NULL;
	delete[] Candidates; Candidates = NULL;
	return;
}